@@ -1,4 +1,4 @@
-import contextlib
+from __future__ import absolute_import
 import netrc
 import os
 import warnings
@@ -6,22 +6,38 @@ from collections import namedtuple
 
 import click
 import six
-import ruamel.yaml as yaml
+import yaml
 
+from shub import DOCS_LINK, CONFIG_DOCS_LINK
 from shub.exceptions import (BadParameterException, BadConfigException,
                              ConfigParseException, MissingAuthException,
-                             NotFoundException)
+                             NotFoundException, ShubDeprecationWarning,
+                             print_warning)
 from shub.utils import (closest_file, get_scrapycfg_targets, get_sources,
-                        pwd_hg_version, pwd_git_version, pwd_version)
+                        pwd_hg_version, pwd_git_version, pwd_version,
+                        update_yaml_dict)
 
-
-GLOBAL_SCRAPINGHUB_YML_PATH = os.path.expanduser("~/.scrapinghub.yml")
+APIKEY_SHOW_N_CHARS = 6
+SH_IMAGES_REGISTRY = 'images.scrapinghub.com'
+SH_IMAGES_REPOSITORY = SH_IMAGES_REGISTRY + '/project/{project}'
+GLOBAL_SCRAPINGHUB_YML_PATH = os.path.expanduser(
+    os.environ.get('SHUB_GLOBAL_CONFIG', '~/.scrapinghub.yml')
+)
 NETRC_PATH = os.path.expanduser('~/_netrc' if os.name == 'nt' else '~/.netrc')
 
 
 class ShubConfig(object):
 
-    DEFAULT_ENDPOINT = 'https://dash.scrapinghub.com/api/'
+    DEFAULT_ENDPOINT = 'https://app.scrapinghub.com/api/'
+
+    # Dictionary option name: Shortcut to set 'default' key
+    SHORTCUTS = {
+        'projects': 'project',
+        'endpoints': 'endpoint',
+        'apikeys': 'apikey',
+        'stacks': 'stack',
+        'images': 'image',
+    }
 
     def __init__(self):
         self.projects = {}
@@ -32,21 +48,83 @@ class ShubConfig(object):
         self.version = 'AUTO'
         self.stacks = {}
         self.requirements_file = None
+        self.eggs = []
+        self.images = {}
+
+    def _check_endpoints(self):
+        """Check the endpoints. Send warnings if necessary."""
+        for endpoint, url in self.endpoints.items():
+            parsed = six.moves.urllib.parse.urlparse(url)
+            if parsed.netloc == 'staging.scrapinghub.com':
+                self.endpoints[endpoint] = six.moves.urllib.parse.urlunparse(
+                    parsed._replace(netloc='app.scrapinghub.com')
+                )
+                click.echo(
+                    'WARNING: Endpoint "%s" is still using %s which has been '
+                    'obsoleted. shub has updated it to app.scrapinghub.com '
+                    'for this time only. Please update your configuration.' % (
+                        endpoint, parsed.netloc,
+                    ),
+                    err=True
+                )
+            if parsed.scheme == 'http':
+                print_warning(
+                    'Endpoint "%s" is still using HTTP. '
+                    'Please change it to HTTPS if possible.' % endpoint
+                )
 
     def load(self, stream):
         """Load Scrapinghub configuration from stream."""
+        # flag to mark if images/default was used in the config file
+        check_default_image_scope = False
         try:
             yaml_cfg = yaml.safe_load(stream)
             if not yaml_cfg:
                 return
-            for option in ('projects', 'endpoints', 'apikeys', 'stacks'):
-                getattr(self, option).update(yaml_cfg.get(option, {}))
+            for option, shortcut in self.SHORTCUTS.items():
+                option_conf = getattr(self, option)
+                yaml_option_conf = yaml_cfg.get(option, {})
+                option_conf.update(yaml_option_conf)
+                if option == 'images' and yaml_option_conf:
+                    print_warning(
+                        "Images section is deprecated, please replace it with "
+                        "global `image` setting or define `image` setting for "
+                        "the project.\n  Check for additional details in {}."
+                        .format(CONFIG_DOCS_LINK),
+                        category=ShubDeprecationWarning
+                    )
+                    if 'default' in yaml_option_conf:
+                        check_default_image_scope = True
+                if shortcut in yaml_cfg:
+                    # We explicitly check yaml_option_conf and not option_conf.
+                    # It is okay to set conflicting defaults if they are in
+                    # different files (b/c then one of these will have
+                    # priority)
+                    if 'default' in yaml_option_conf:
+                        raise BadConfigException(
+                            "You cannot specify both '%s' and a 'default' key "
+                            "for '%s' in the same file" % (shortcut,  option))
+                    option_conf['default'] = yaml_cfg[shortcut]
             self.version = yaml_cfg.get('version', self.version)
-            if 'requirements_file' in yaml_cfg:
-                self.requirements_file = yaml_cfg['requirements_file']
+            self.requirements_file = yaml_cfg.get('requirements_file',
+                                                  self.requirements_file)
+            self.requirements_file = yaml_cfg.get('requirements', {}).get(
+                'file', self.requirements_file)
+            self.eggs = yaml_cfg.get('requirements', {}).get('eggs', self.eggs)
         except (yaml.YAMLError, AttributeError):
             # AttributeError: stream is valid YAML but not dictionary-like
             raise ConfigParseException
+        # fail if `projects` section has keys not found in `images`
+        if (check_default_image_scope and
+                not set(self.projects).issubset(set(self.images))):
+            raise BadConfigException(
+                    "Found ambigious configuration: default image has global "
+                    "scope now, but some projects were not using custom "
+                    "images and can be broken now. Please fix your config by "
+                    "replacing `images` section with `image` settings. Check "
+                    "for additional details in {}".format(CONFIG_DOCS_LINK)
+                )
+        self._check_endpoints()
 
     def load_file(self, filename):
         """Load Scrapinghub configuration from YAML file. """
@@ -87,30 +165,49 @@ class ShubConfig(object):
         del targets['default']
         for tname, t in six.iteritems(targets):
             self._load_scrapycfg_target(tname, t)
+        self._check_endpoints()
+
+    def save(self, path=None, options=None):
+        def _project_id_as_int(project):
+            """Copy project and return it with the ID casted to int to make
+            sure it is exported as "123" and not "'123'"
+            """
+            try:
+                if isinstance(project, dict):
+                    project = project.copy()
+                    project['id'] = int(project['id'])
+                else:
+                    project = int(project)
+            except ValueError:
+                # Happens when project ID contains endpoint (e.g. vagrant/123),
+                # in that case it will be exported without quotation marks
+                # anyway
+                pass
+            return project
 
-    def save(self, path=None):
         with update_yaml_dict(path) as yml:
-            yml['projects'] = self.projects
-            # Write "123" instead of "'123'"
-            for target, project in yml['projects'].iteritems():
-                try:
-                    if isinstance(project, dict):
-                        project['id'] = int(project['id'])
-                    else:
-                        yml['projects'][target] = int(project)
-                except Exception:
-                    pass
-            yml['endpoints'] = self.endpoints
-            yml['apikeys'] = self.apikeys
-            yml['version'] = self.version
-            yml['stacks'] = self.stacks
+            options = options or list(self.SHORTCUTS)
+            for option in options:
+                shortcut = self.SHORTCUTS[option]
+                conf = getattr(self, option)
+                if option == 'endpoints' and conf == ShubConfig().endpoints:
+                    # Don't write default endpoint
+                    continue
+                elif option == 'projects':
+                    conf = {k: _project_id_as_int(v) for k, v in conf.items()}
+                if list(conf.keys()) == ['default']:
+                    yml[shortcut] = conf['default']
+                    yml.pop(option, None)
+                else:
+                    yml[option] = conf
+                    yml.pop(shortcut, None)
+            if self.version != 'AUTO':
+                yml['version'] = self.version
+            if self.eggs:
+                yml.setdefault('requirements', {})['eggs'] = self.eggs
             if self.requirements_file:
-                yml['requirements_file'] = self.requirements_file
-            # Don't write defaults
-            if self.endpoints['default'] == ShubConfig.DEFAULT_ENDPOINT:
-                del yml['endpoints']['default']
-            if self.version == 'AUTO':
-                del yml['version']
+                yml.setdefault('requirements', {})['file'] = (
+                    self.requirements_file)
 
     @property
     def normalized_projects(self):
@@ -119,7 +216,7 @@ class ShubConfig(object):
         that have at least the keys ``id``, ``endpoint``, and ``apikey``.
         """
         projects = self.projects.copy()
-        for target, proj in projects.items():
+        for target, proj in list(projects.items()):
             if not isinstance(proj, dict):
                 proj = {'id': proj}
                 projects[target] = proj
@@ -201,10 +298,32 @@ class ShubConfig(object):
             apikey=apikey,
             stack=(self.stacks.get(proj['stack'], proj['stack'])
                    if 'stack' in proj else self.stacks.get('default')),
+            image=self._select_image_for_project(target, proj),
             requirements_file=self.requirements_file,
             version=self.get_version(),
+            eggs=self.eggs,
         )
 
+    def _select_image_for_project(self, target, project):
+        """Helper to select image for a project (or its target).
+
+        The select logic is the following:
+        - image defined per project has highest priority,
+        - images section is marked as deprecated, but still in force:
+          if target is defined in images - use a corresponding image,
+        - if image is not defined, but there's a default image set by
+          image or images/default settings - use it.
+
+        The function responds with a custom image name string
+        (or None/False meaning regular stack-based deploy).
+        """
+        image = project.get('image', self.images.get(
+            target, self.images.get('default')))
+        # aliases to use internal scrapinghub registry as image storage
+        if image is True or image == 'scrapinghub':
+            image = SH_IMAGES_REPOSITORY.format(project=project['id'])
+        return image
+
     def get_target(self, target, auth_required=True):
         """Return (project_id, endpoint, apikey) for given target."""
         warnings.warn("get_target is deprecated, use get_target_conf instead")
@@ -222,11 +341,64 @@ class ShubConfig(object):
         return self.get_target_conf(target, auth_required=False).endpoint
 
     def get_apikey(self, target, required=True):
-        return self.get_target_conf(target, auth_required=required).apikey
+        apikey = self.get_target_conf(target, auth_required=required).apikey
+        return getattr(apikey, 'value', apikey)
+
+    def get_image(self, target):
+        """Return image for a given target."""
+        target_conf = self.get_target_conf(target, auth_required=False)
+        project, image = target_conf.project_id, target_conf.image
+        if image is None:
+            raise NotFoundException(
+                "Could not find image for project '{}'. Please define it "
+                "in your scrapinghub.yml.".format(target))
+        elif image is False:
+            raise BadConfigException(
+                "Using custom images is disabled for the project '{}'. "
+                "Please enable it in your scrapinghub.yml.".format(target))
+        elif target_conf.stack:
+            raise BadConfigException(
+                "Ambiguous configuration: There is both a custom image and a "
+                "stack configured for project '{}'. Please see {} for "
+                "information on how to configure both custom image-based and "
+                "stack-based projects.".format(target, CONFIG_DOCS_LINK))
+        default_image = SH_IMAGES_REPOSITORY.format(project=project)
+        if image.startswith(SH_IMAGES_REGISTRY) and image != default_image:
+            raise BadConfigException(
+                "Found wrong SH repository for project '{}': expected {}.\n  "
+                "Please use aliases `True` or `scrapinghub` to fix it in your "
+                "config.".format(target, default_image))
+        return image
 
 
-Target = namedtuple('Target', ['project_id', 'endpoint', 'apikey', 'stack',
-                               'requirements_file', 'version'])
+_Target = namedtuple('Target', ['project_id', 'endpoint', 'apikey', 'stack',
+                                'image', 'requirements_file', 'version',
+                                'eggs'])
+
+
+class APIkey(str):
+
+    def __new__(cls, *args, **kwargs):
+        cls._inst = super(APIkey, cls).__new__(cls, *args, **kwargs)
+        return cls._inst
+
+    def __init__(self, value=None):
+        self.value = value
+
+    def __repr__(self):
+        if not self.value:
+            return ''
+        visible_chars = APIKEY_SHOW_N_CHARS
+        return (self.value[:visible_chars] +
+                'X' * max(len(self.value) - visible_chars, 0))
+
+
+class Target(_Target):
+
+    def __new__(cls, project_id, endpoint, apikey, *args, **kwargs):
+        cls._inst = super(Target, cls).__new__(cls, project_id, endpoint,
+                                               APIkey(apikey), *args, **kwargs)
+        return cls._inst
 
 
 MIGRATION_BANNER = """
@@ -249,12 +421,12 @@ But no worries, shub has automatically migrated your global settings to
 ~/.scrapinghub.yml, and will also automatically migrate your project settings
 when you run a command within a Scrapy project.
 
-Visit http://doc.scrapinghub.com/shub.html for more information on the new
-configuration format and its benefits.
+Visit {docs_link} for more information on the new configuration format and
+its benefits.
 
 Happy scraping!
 -------------------------------------------------------------------------------
-"""
+""".format(docs_link=DOCS_LINK)
 
 
 def _migrate_to_global_scrapinghub_yml():
@@ -267,7 +439,7 @@ def _migrate_to_global_scrapinghub_yml():
         netrc_key = None
     if netrc_key:
         conf.apikeys['default'] = netrc_key
-    conf.save()
+    conf.save(GLOBAL_SCRAPINGHUB_YML_PATH)
     default_conf = ShubConfig()
     migrated_data = any(getattr(conf, attr) != getattr(default_conf, attr)
                         for attr in ('projects', 'endpoints', 'apikeys',
@@ -279,19 +451,19 @@ def _migrate_to_global_scrapinghub_yml():
 PROJECT_MIGRATION_OK_BANNER = """
 INFO: Your deploy configuration has been migrated to scrapinghub.yml.
 shub will no longer read from scrapy.cfg (but Scrapy will, so don't delete it).
-Visit http://doc.scrapinghub.com/shub.html for more information.
-"""
+Visit {docs_link} for more information.
+""".format(docs_link=DOCS_LINK)
 
 
 PROJECT_MIGRATION_FAILED_BANNER = """
 WARNING: shub failed to convert your scrapy.cfg to scrapinghub.yml. Please
-visit http://doc.scrapinghub.com/shub.html for help on how to use the new
-configuration format. We would be grateful if you could also file a bug report
-at https://github.com/scrapinghub/shub/issues
+visit {docs_link} for help on how to use the new configuration format. We
+would be grateful if you could also file a bug report at
+https://github.com/scrapinghub/shub/issues
 
 For now, shub fell back to reading from scrapy.cfg, everything should work as
 expected.
-"""
+""".format(docs_link=DOCS_LINK)
 
 
 def _migrate_and_load_scrapy_cfg(conf):
@@ -338,35 +510,6 @@ def load_shub_config(load_global=True, load_local=True, load_env=True):
     return conf
 
 
-@contextlib.contextmanager
-def update_yaml_dict(conf_path=None):
-    """
-    Context manager for updating a YAML file while preserving key ordering and
-    comments.
-    """
-    conf_path = conf_path or GLOBAL_SCRAPINGHUB_YML_PATH
-    dumper = yaml.RoundTripDumper
-    try:
-        with open(conf_path, 'r') as f:
-            conf = yaml.load(f, yaml.RoundTripLoader) or {}
-    except IOError as e:
-        if e.errno != 2:
-            raise
-        conf = {}
-        # Use alphabetic order when creating files
-        dumper = yaml.Dumper
-    # Code inside context manager is executed after this yield
-    yield conf
-    # Avoid writing "key: {}"
-    for key in conf.keys():
-        if conf[key] == {}:
-            del conf[key]
-    with open(conf_path, 'w') as f:
-        # Avoid writing "{}"
-        if conf:
-            yaml.dump(conf, f, default_flow_style=False, Dumper=dumper)
-
-
 def get_target(target, auth_required=True):
     """Load shub configuration and return target."""
     conf = load_shub_config()
@@ -383,3 +526,13 @@ def get_version():
     """Load shub configuratoin and return version."""
     conf = load_shub_config()
     return conf.get_version()
+
+
+def list_targets_callback(ctx, param, value):
+    """Click option callback to get targets from config, print it and exit."""
+    if not value:
+        return
+    conf = load_shub_config()
+    for name in conf.projects:
+        click.echo(name)
+    ctx.exit()
